/*
Copyright 1990-2001 Sun Microsystems, Inc. All Rights Reserved.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions: The above copyright notice and this
permission notice shall be included in all copies or substantial
portions of the Software.


THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE OPEN GROUP OR SUN MICROSYSTEMS, INC. BE LIABLE
FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH
THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE EVEN IF
ADVISED IN ADVANCE OF THE POSSIBILITY OF SUCH DAMAGES.


Except as contained in this notice, the names of The Open Group and/or
Sun Microsystems, Inc. shall not be used in advertising or otherwise to
promote the sale, use or other dealings in this Software without prior
written authorization from The Open Group and/or Sun Microsystems,
Inc., as applicable.


X Window System is a trademark of The Open Group

OSF/1, OSF/Motif and Motif are registered trademarks, and OSF, the OSF
logo, LBX, X Window System, and Xinerama are trademarks of the Open
Group. All other trademarks and registered trademarks mentioned herein
are the property of their respective owners. No right, title or
interest in or to any trademark, service mark, logo or trade name of
Sun Microsystems, Inc. or its licensors is granted.

*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdlib.h>
#include <string.h>

#ifdef HAVE_CRYPT
#include <crypt.h>
#endif

#ifdef HAVE_PWENT
#include <sys/types.h>
#include <pwd.h>
#endif

#include "iwrap.h"
#include "UserAuth.hh"

int
UserAuth::
adjust_pue_slot_size(int num)
{
  alloced_ue_num = num;
  pue = (userentry*) realloc(pue, sizeof(userentry) * alloced_ue_num);
  if (!pue) return 0;
  return 1;
}

int
UserAuth::
check_password_with_crypt(char *pwd, char *enc_pwd)
{
  char *p;
#ifdef HAVE_CRYPT
  if (!(p = crypt(pwd, enc_pwd)))
    return 0;
  if (!strcmp(p, enc_pwd))
    return 1;
  return 0;
#else
  return 0;
#endif
}


Authenticator::access_type
UserAuth::
get_internal_userentry(char *user, char *password)
{
  int i;
  char *p;
  userentry *pe;

  pe = pue;
  for (i = 0;i < ue_num;i++, pe++)
    {
      if (strcmp(pe->username, user) != 0) continue;
      if (pe->at == Authenticator::PASSWORD)
	{
	  if (password
	      && pe->password
	      && (check_password_with_crypt(password, pe->password)))
	    return Authenticator::PERMIT;
	  return Authenticator::DENY;
	}
      else
	return pe->at;
    }
  return Authenticator::UNKNOWN;
}

int
UserAuth::
check_password(char *user, char *password)
{
  Authenticator::access_type at;

  at = get_internal_userentry(user, password);
  if (at == Authenticator::PERMIT) return 1;
  else if (at == Authenticator::DENY) return 0;

#ifdef HAVE_PAM
  if (sysuser_at == Authenticator::PERMIT)
    return authenticate_with_pam(user, NULL,
				 get_command_name(),
				 get_from_hostname());
  else if (password && (sysuser_at == Authenticator::PASSWORD))
    return authenticate_with_pam(user, password,
				 get_command_name(),
				 get_from_hostname());
  else
    return 0;
#else
#ifdef HAVE_PWENT
  {
    struct passwd *pwd;
    if (!(pwd = getpwnam(user))) return 0;
    if (sysuser_at == Authenticator::PERMIT) return 1;
    if (sysuser_at == Authenticator::PASSWORD)
      {
	if (password
	    &&(check_password_with_crypt(password, pwd->pw_passwd)))
	  return 1;
	return 0;
      }
  }
#endif
#endif
  return 0;
}

int
UserAuth::
add_user(char *user, char *password, access_type at)
{
  char *u1, *p1;

  if (alloced_ue_num <= ue_num)
    {
      if (!adjust_pue_slot_size((alloced_ue_num + 1) * 2))
	return 0;
    }

  u1= strdup(user);
  if (!u1) return 0;
  if (password)
    {
      p1 = strdup(password);
      if (!p1)
	{
	  free(u1);
	  return 0;
	}
    }
  else
    p1 = NULL;
  pue[ue_num].username = u1;
  pue[ue_num].password = p1;
  pue[ue_num].at = at;
  ue_num++;

  return 1;
}

UserAuth::
UserAuth()
  : Authenticator()
{
  pue = NULL;
  ue_num = alloced_ue_num = 0;
  sysuser_at = Authenticator::DENY;
}

UserAuth::
~UserAuth()
{
  int i;
  for (i = 0;i < ue_num;i++)
    {
      if (pue[i].username) free(pue[i].username);
      if (pue[i].password) free(pue[i].password);
    }
  if (pue) free(pue);
}

